@@ -8,7 +8,6 @@ import traceback
 from glob import glob
 
 from hdfs import InsecureClient, HdfsError
-from .sqlcliexception import SQLCliException
 
 
 class HDFSWrapperException(Exception):
@@ -28,13 +27,13 @@ class HDFSWrapper(object):
     def HDFS_makedirs(self, hdfs_path):
         """ 创建目录 """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
         self.__m_HDFS_Handler__.makedirs(os.path.join(self.__m_HDFS_WebFSDir__, hdfs_path).replace('\\', '/'))
 
     def HDFS_setPermission(self, hdfs_path, permission):
         """ 修改指定文件的权限信息 """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
         m_hdfs_filepath = os.path.dirname(hdfs_path)
         m_hdfs_filename = os.path.basename(hdfs_path)
         self.__m_HDFS_Handler__.set_permission(
@@ -66,17 +65,32 @@ class HDFSWrapper(object):
     def HDFS_status(self, hdfs_path=""):
         """ 返回目录下的文件 """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
 
         m_ReturnList = []
         m_Status = self.__m_HDFS_Handler__.status(hdfs_path)
+        if m_Status['type'].upper() == 'DIRECTORY':
+            # HDFS CLI对于目录的大小总是返回0， 所以这里遍历所有的目录来获得准确的目录大小
+            m_FileList = self.HDFS_list(hdfs_path, recusive=True)
+            m_FileSize = 0
+            for m_File in m_FileList:
+                m_FileSize = m_FileSize + int(m_File[1]['length'])
+            if m_FileSize > 1024*1024*1024:
+                m_FileSize = str(round(float(m_FileSize) / (1024*1024*1024))) + 'G'
+            elif m_FileSize > 1024*1024:
+                m_FileSize = str(round(float(m_FileSize) / (1024*1024))) + 'M'
+            elif m_FileSize > 1024:
+                m_FileSize = str(round(float(m_FileSize) / 1024)) + 'K'
+            else:
+                m_FileSize = str(m_FileSize)
+            m_Status['length'] = m_FileSize
         m_ReturnList.append((hdfs_path, m_Status))
         return m_ReturnList
 
     def HDFS_list(self, hdfs_path="", recusive=False):
         """ 返回目录下的文件 """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
 
         m_ReturnList = []
         if not recusive:
@@ -98,7 +112,7 @@ class HDFSWrapper(object):
     def HDFS_Download(self, hdfs_path="", local_path="", recusive=False):
         """ 从hdfs获取文件到本地 """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
 
         # 如果本地没有对应目录，且local_path传递的是一个目录，则建立目录
         m_LocalPath = local_path
@@ -113,7 +127,7 @@ class HDFSWrapper(object):
     def HDFS_Upload(self, local_path, hdfs_path=""):
         """ 上传文件到hdfs """
         if self.__m_HDFS_Handler__ is None:
-            raise HDFSWrapperException("HDFS not connected. Please connect it frist.")
+            raise HDFSWrapperException("HDFS not connected. Please connect it first.")
 
         for file in glob(local_path):
             if hdfs_path == "":
@@ -330,4 +344,4 @@ class HDFSWrapper(object):
             if "SQLCLI_DEBUG" in os.environ:
                 print('traceback.print_exc():\n%s' % traceback.print_exc())
                 print('traceback.format_exc():\n%s' % traceback.format_exc())
-            raise SQLCliException(he.message)
+            return None, None, None, None, he.message
