@@ -41,9 +41,12 @@ class SQLExecute(object):
         # 程序Echo输出句柄
         self.echofile = None
 
-        # Scenario信息，如果当前SQL指定，则重复上一个的Scenario信息
+        # Scenario名称，如果当前SQL未指定，则重复上一个SQL的Scenario信息
         self.SQLScenario = ''
 
+        # Scenario的优先级，如果当前SQL未指定，则重复上一个SQL的Scenario信息
+        self.SQLPriority = ''
+
         # Transaction信息
         self.SQLTransactionName = ""
 
@@ -70,7 +73,7 @@ class SQLExecute(object):
             pass
         if obj is None:
             if "SQLCLI_DEBUG" in os.environ:
-                click.secho("JQ Parse Error: obj is None")
+                click.secho("[DEBUG] JQ Parse Error: obj is None")
             return "****"
         try:
             obj = json.loads(obj) if isinstance(obj, str) else obj
@@ -80,7 +83,7 @@ class SQLExecute(object):
                     continue
                 if obj is None:
                     if "SQLCLI_DEBUG" in os.environ:
-                        click.secho("JQ Parse Error: obj is none")
+                        click.secho("[DEBUG] JQ Parse Error: obj is none")
                     return "****"
                 if isinstance(obj, (list, tuple, str)):
                     if im.startswith('[') and im.endswith(']'):
@@ -109,7 +112,7 @@ class SQLExecute(object):
                                                                cls=DecimalEncoder)
         except (IndexError, KeyError, ValueError) as je:
             if "SQLCLI_DEBUG" in os.environ:
-                click.secho("JQ Parse Error: " + repr(je))
+                click.secho("[DEBUG] JQ Parse Error: " + repr(je))
             return "****"
 
     def run(self, statement, p_sqlscript=None):
@@ -145,24 +148,31 @@ class SQLExecute(object):
         }
 
         支持的SQLHint包括:
-        -- [Hint]  order           -- SQLCli将会把随后的SQL语句进行排序输出，原程序的输出顺序被忽略
-        -- [Hint]  scenario:XXXX   -- 相关SQL的场景ID，仅仅作为日志信息供查看
+        -- [Hint]  order              -- SQLCli将会把随后的SQL语句进行排序输出，原程序的输出顺序被忽略
+        -- [Hint]  scenario:XXXX      -- 相关SQL的场景名称
+        -- [Hint]  scenario:XXXX:P1   -- 相关SQL的场景名称以及SQL优先级，如果应用程序指定了限制的优先级，则只运行指定的优先级脚本
         -- .....
         """
-        m_SQL_Status = 0             # SQL 运行结果， 0 成功， 1 失败
-        m_SQL_ErrorMessage = ""      # 错误日志信息
-
         # Remove spaces and EOL
         statement = statement.strip()
         if not statement:  # Empty string
             return
 
+        # 优先级
+        m_Priorites = []
+        if self.SQLOptions.get("PRIORITY") != "":
+            for m_Priority in self.SQLOptions.get("PRIORITY").split(','):
+                m_Priorites.append(m_Priority.strip().upper())
+
         # 分析SQL语句
+        m_HasOutputScenarioSkipInfo = False
         (ret_bSQLCompleted, ret_SQLSplitResults,
          ret_SQLSplitResultsWithComments, ret_SQLHints) = SQLAnalyze(statement)
-        for m_nPos in range(0, len(ret_SQLSplitResults)):
-            m_raw_sql = ret_SQLSplitResults[m_nPos]                  # 记录原始SQL
 
+        for m_nPos in range(0, len(ret_SQLSplitResults)):
+            m_raw_sql = ret_SQLSplitResults[m_nPos]                 # 记录原始SQL
+            m_SQL_ErrorMessage = ""                                 # 错误日志信息
+            m_SQL_Status = 0                                        # SQL 运行结果， 0 成功， 1 失败
             # 如果当前是在回显一个文件，则不再做任何处理，直接返回
             if self.echofile is not None and \
                     not re.match(r'echo\s+off', m_raw_sql, re.IGNORECASE):
@@ -181,9 +191,31 @@ class SQLExecute(object):
             m_SQLHint = ret_SQLHints[m_nPos]                         # SQL提示信息，其中Scenario用作日志处理
             if "SCENARIO" in m_SQLHint.keys():
                 if m_SQLHint['SCENARIO'].strip().upper() == "END":
+                    m_HasOutputScenarioSkipInfo = False
                     self.SQLScenario = ""
+                    self.SQLPriority = ""
                 else:
                     self.SQLScenario = m_SQLHint['SCENARIO']
+                    m_HasOutputScenarioSkipInfo = False
+                    if "PRIORITY" in m_SQLHint.keys():
+                        self.SQLPriority = m_SQLHint['PRIORITY']
+
+            # 检查优先级, 如果定义了优先级，但不符合定义要求，则直接跳过这个语句
+            if self.SQLPriority != "" and len(m_Priorites) != 0:
+                if self.SQLPriority not in m_Priorites:
+                    # 由于优先级原因，这个SQL不会被执行，直接返回
+                    # Scenario的第一个语句打印Skip信息
+                    if not m_HasOutputScenarioSkipInfo:
+                        yield {
+                            "type": "result",
+                            "title": None,
+                            "rows": None,
+                            "headers": None,
+                            "columntypes": None,
+                            "status": "Skip scenario [" + self.SQLScenario + "] due to priority set."
+                        }
+                        m_HasOutputScenarioSkipInfo = True
+                    continue
 
             # 记录包含有注释信息的SQL
             m_FormattedSQL = SQLFormatWithPrefix(m_CommentSQL)
@@ -213,8 +245,7 @@ class SQLExecute(object):
             # 1. ${LastSQLResult(.*)}       # .* JQ Parse Pattern
             # 2. ${var}
             #    用户定义的变量
-
-            matchObj = re.search(r"\${LastSQLResult\((.*)\)}",
+            matchObj = re.search(r"\${LastSQLResult\((.*?)\)}",
                                  sql, re.IGNORECASE | re.DOTALL)
             if matchObj:
                 m_Searched = matchObj.group(0)
@@ -283,16 +314,21 @@ class SQLExecute(object):
             except ValueError:
                 m_SQLTimeOut = -1
                 if "SQLCLI_DEBUG" in os.environ:
-                    yield {"type": "error", "message": "SQLCLI-0000: Not a valid SQLTimeOut Setting, Ignore this."}
+                    yield {"type": "error",
+                           "message": "[DEBUG] SQLCLI-0000: Not a valid SQLTimeOut Setting, Ignore this."}
             try:
                 m_ScriptTimeOut = int(self.SQLOptions.get("SCRIPT_TIMEOUT"))
             except ValueError:
                 m_ScriptTimeOut = -1
                 if "SQLCLI_DEBUG" in os.environ:
-                    yield {"type": "error", "message": "SQLCLI-0000: Not a valid ScriptTimeOut Setting, Ignore this."}
+                    yield {"type": "error",
+                           "message": "[DEBUG] SQLCLI-0000: Not a valid ScriptTimeOut Setting, Ignore this."}
 
             # 执行SQL
             try:
+                if "SQLCLI_DEBUG" in os.environ:
+                    print("[DEBUG] SQL=[" + str(sql) + "]")
+
                 # 处理超时时间问题
                 if m_ScriptTimeOut > 0:
                     if m_ScriptTimeOut <= time.time() - self.getStartTime():
@@ -306,13 +342,74 @@ class SQLExecute(object):
                         if m_SQLTimeOut <= 0 or (m_ScriptTimeOut - (time.time() - self.getStartTime()) < m_SQLTimeOut):
                             # 脚本时间已经不够，剩下的SQL时间变少
                             m_SQLTimeOut = m_ScriptTimeOut - (time.time() - self.getStartTime())
+
                 # 首先尝试这是一个特殊命令，如果返回CommandNotFound，则认为其是一个标准SQL
                 for m_Result in execute(self.SQLCliHandler,  sql, p_nTimeout=m_SQLTimeOut):
                     # 保存之前的运行结果
                     if "type" not in m_Result.keys():
                         m_Result.update({"type": "result"})
-                    if m_Result["type"] == "result" and sql.startswith("__internal__"):
+                    if m_Result["type"] == "result" and \
+                            (sql.lower().startswith("__internal__") or sql.lower().startswith("loaddriver")):
+                        # 如果存在SQL_LOOP信息，则需要反复执行上一个SQL
+                        if "SQL_LOOP" in m_SQLHint.keys():
+                            if "SQLCLI_DEBUG" in os.environ:
+                                print("[DEBUG] LOOP=" + str(m_SQLHint["SQL_LOOP"]))
+                            # 循环执行SQL列表，构造参数列表
+                            m_LoopTimes = int(m_SQLHint["SQL_LOOP"]["LoopTimes"])
+                            m_LoopInterval = int(m_SQLHint["SQL_LOOP"]["LoopInterval"])
+                            m_LoopUntil = m_SQLHint["SQL_LOOP"]["LoopUntil"]
+                            if m_LoopInterval < 0:
+                                m_LoopTimes = 0
+                                if "SQLCLI_DEBUG" in os.environ:
+                                    raise SQLCliException(
+                                        "SQLLoop Hint Error, Unexpected LoopInterval: " + str(m_LoopInterval))
+                            if m_LoopTimes < 0:
+                                if "SQLCLI_DEBUG" in os.environ:
+                                    raise SQLCliException(
+                                        "SQLLoop Hint Error, Unexpected LoopTime: " + str(m_LoopTimes))
+
+                            # 保存Silent设置
+                            m_OldSilentMode = self.SQLOptions.get("SILENT")
+                            m_OldTimingMode = self.SQLOptions.get("TIMING")
+                            m_OldTimeMode = self.SQLOptions.get("TIME")
+                            self.SQLOptions.set("SILENT", "ON")
+                            self.SQLOptions.set("TIMING", "OFF")
+                            self.SQLOptions.set("TIME", "OFF")
+                            for m_nLoopPos in range(1, m_LoopTimes):
+                                # 检查Until条件，如果达到Until条件，退出
+                                m_AssertSuccessful = False
+                                for m_SQLResult in \
+                                        self.run("__internal__ test assert " + m_LoopUntil):
+                                    if m_SQLResult["type"] == "result":
+                                        if m_SQLResult["status"].startswith("Assert Successful"):
+                                            m_AssertSuccessful = True
+                                        break
+                                if m_AssertSuccessful:
+                                    break
+                                else:
+                                    # 测试失败, 等待一段时间后，开始下一次检查
+                                    time.sleep(m_LoopInterval)
+                                    if "SQLCLI_DEBUG" in os.environ:
+                                        print("[DEBUG] SQL(LOOP " + str(m_nLoopPos) + ")=[" + str(sql) + "]")
+                                    for m_SQLResult in self.run(sql):
+                                        # 最后一次执行的结果将被传递到外层，作为SQL返回结果
+                                        if m_SQLResult["type"] == "result":
+                                            m_SQL_Status = 0
+                                            m_Result["title"] = m_SQLResult["title"]
+                                            m_Result["rows"] = m_SQLResult["rows"]
+                                            m_Result["headers"] = m_SQLResult["headers"]
+                                            m_Result["columntypes"] = m_SQLResult["columntypes"]
+                                            m_Result["status"] = m_SQLResult["status"]
+                                        if m_SQLResult["type"] == "error":
+                                            m_SQL_Status = 1
+                                            m_SQL_ErrorMessage = m_SQLResult["message"]
+                            self.SQLOptions.set("TIME", m_OldTimeMode)
+                            self.SQLOptions.set("TIMING", m_OldTimingMode)
+                            self.SQLOptions.set("SILENT", m_OldSilentMode)
+
+                        # 保存之前的运行结果
                         result = m_Result["rows"]
+
                         # 如果Hints中有order字样，对结果进行排序后再输出
                         if "Order" in m_SQLHint.keys() and result is not None:
                             for i in range(1, len(result)):
@@ -325,7 +422,8 @@ class SQLExecute(object):
                             for m_SQLFilter in m_SQLHint["LogFilter"]:
                                 for item in result[:]:
                                     if "SQLCLI_DEBUG" in os.environ:
-                                        print("Apply Filter: " + str(''.join(str(item))) + " with " + m_SQLFilter)
+                                        print("[DEBUG] Apply Filter: " + str(''.join(str(item))) +
+                                              " with " + m_SQLFilter)
                                     if re.match(m_SQLFilter, ''.join(str(item)), re.IGNORECASE):
                                         result.remove(item)
                                         continue
@@ -345,7 +443,7 @@ class SQLExecute(object):
                                         m_SQLMaskPattern = m_SQLMask[0]
                                         m_SQLMaskTarget = m_SQLMask[1]
                                         if "SQLCLI_DEBUG" in os.environ:
-                                            print("Apply Mask: " + m_Output +
+                                            print("[DEBUG] Apply Mask: " + m_Output +
                                                   " with " + m_SQLMaskPattern + "=>" + m_SQLMaskTarget)
                                         try:
                                             m_NewOutput = re.sub(m_SQLMaskPattern, m_SQLMaskTarget, m_Output,
@@ -355,11 +453,13 @@ class SQLExecute(object):
                                                 m_Output = m_NewOutput
                                         except re.error:
                                             if "SQLCLI_DEBUG" in os.environ:
-                                                print('traceback.print_exc():\n%s' % traceback.print_exc())
-                                                print('traceback.format_exc():\n%s' % traceback.format_exc())
+                                                print('[DEBUG] traceback.print_exc():\n%s' % traceback.print_exc())
+                                                print('[DEBUG] traceback.format_exc():\n%s' % traceback.format_exc())
                                     else:
                                         if "SQLCLI_DEBUG" in os.environ:
-                                            print("LogMask Hint Error: " + m_SQLHint["LogMask"])
+                                            print("[DEBUG] LogMask Hint Error: " + m_SQLHint["LogMask"])
+
+                        # 返回运行结果
                         m_Result["rows"] = result
                         if m_Result["rows"] is None:
                             m_Rows = 0
@@ -408,8 +508,6 @@ class SQLExecute(object):
                 # 执行正常的SQL语句
                 if self.cur is not None:
                     try:
-                        if "SQLCLI_DEBUG" in os.environ:
-                            print("DEBUG-SQL=[" + str(sql) + "]")
                         # 处理超时时间问题
                         if m_ScriptTimeOut > 0:
                             if m_ScriptTimeOut <= time.time() - self.getStartTime():
@@ -451,11 +549,11 @@ class SQLExecute(object):
                                 self.get_result(self.cur, rowcount)
                             rowcount = m_FetchedRows
                             if "SQLCLI_DEBUG" in os.environ:
-                                print("!!DEBUG!!  headers=" + str(headers))
+                                print("[DEBUG] headers=" + str(headers))
                                 if result is not None:
                                     for m_RowPos in range(0, len(result)):
                                         for m_CellPos in range(0, len(result[m_RowPos])):
-                                            print("Cell[" + str(m_RowPos) + ":" +
+                                            print("[DEBUG] Cell[" + str(m_RowPos) + ":" +
                                                   str(m_CellPos) + "]=[" + str(result[m_RowPos][m_CellPos]) + "]")
 
                             # 如果Hints中有order字样，对结果进行排序后再输出
@@ -470,7 +568,8 @@ class SQLExecute(object):
                                 for m_SQLFilter in m_SQLHint["LogFilter"]:
                                     for item in result[:]:
                                         if "SQLCLI_DEBUG" in os.environ:
-                                            print("Apply Filter: " + str(''.join(str(item))) + " with " + m_SQLFilter)
+                                            print("[DEBUG] Apply Filter: " + str(''.join(str(item))) +
+                                                  " with " + m_SQLFilter)
                                         if re.match(m_SQLFilter, ''.join(str(item)), re.IGNORECASE):
                                             result.remove(item)
                                             continue
@@ -490,7 +589,7 @@ class SQLExecute(object):
                                             m_SQLMaskPattern = m_SQLMask[0]
                                             m_SQLMaskTarget = m_SQLMask[1]
                                             if "SQLCLI_DEBUG" in os.environ:
-                                                print("Apply Mask: " + m_Output +
+                                                print("[DEBUG] Apply Mask: " + m_Output +
                                                       " with " + m_SQLMaskPattern + "=>" + m_SQLMaskTarget)
                                             try:
                                                 m_NewOutput = re.sub(m_SQLMaskPattern, m_SQLMaskTarget, m_Output,
@@ -500,11 +599,13 @@ class SQLExecute(object):
                                                     m_Output = m_NewOutput
                                             except re.error:
                                                 if "SQLCLI_DEBUG" in os.environ:
-                                                    print('traceback.print_exc():\n%s' % traceback.print_exc())
-                                                    print('traceback.format_exc():\n%s' % traceback.format_exc())
+                                                    print('[DEBUG] traceback.print_exc():\n%s'
+                                                          % traceback.print_exc())
+                                                    print('[DEBUG] traceback.format_exc():\n%s'
+                                                          % traceback.format_exc())
                                         else:
                                             if "SQLCLI_DEBUG" in os.environ:
-                                                print("LogMask Hint Error: " + m_SQLHint["LogMask"])
+                                                print("[DEBUG] LogMask Hint Error: " + m_SQLHint["LogMask"])
 
                             # 保存之前的运行结果
                             if result is None:
@@ -521,7 +622,7 @@ class SQLExecute(object):
                             # 如果存在SQL_LOOP信息，则需要反复执行上一个SQL
                             if "SQL_LOOP" in m_SQLHint.keys():
                                 if "SQLCLI_DEBUG" in os.environ:
-                                    print("SQL_LOOP=" + str(m_SQLHint["SQL_LOOP"]))
+                                    print("[DEBUG] LOOP=" + str(m_SQLHint["SQL_LOOP"]))
                                 # 循环执行SQL列表，构造参数列表
                                 m_LoopTimes = int(m_SQLHint["SQL_LOOP"]["LoopTimes"])
                                 m_LoopInterval = int(m_SQLHint["SQL_LOOP"]["LoopInterval"])
@@ -557,6 +658,8 @@ class SQLExecute(object):
                                     else:
                                         # 测试失败, 等待一段时间后，开始下一次检查
                                         time.sleep(m_LoopInterval)
+                                        if "SQLCLI_DEBUG" in os.environ:
+                                            print("[DEBUG] SQL(LOOP " + str(m_nLoopPos) + ")=[" + str(sql) + "]")
                                         for m_Result in self.run(sql):
                                             # 最后一次执行的结果将被传递到外层，作为SQL返回结果
                                             if m_Result["type"] == "result":
@@ -724,8 +827,13 @@ class SQLExecute(object):
                     "ErrorMessage": m_SQL_ErrorMessage,
                     "thread_name": self.WorkerName,
                     "Scenario": self.SQLScenario,
+                    "Priority": self.SQLPriority,
                     "Transaction": self.SQLTransactionName
                 }
+        # 当前SQL文件已经执行完毕，清空Scenario，以及Priority
+        if p_sqlscript is not None:
+            self.SQLScenario = ''
+            self.SQLPriority = ''
 
     def get_result(self, cursor, rowcount):
         """
@@ -749,7 +857,7 @@ class SQLExecute(object):
                 return p_column
             elif type(p_column) == datetime.date:
                 m_ColumnFormat = self.SQLOptions.get("DATE_FORMAT")
-                if platform.system().lower() == 'windows':
+                if platform.system().lower() in ['windows', 'darwin']:
                     m_ColumnFormat = m_ColumnFormat.replace("%04Y", "%Y")
                 else:
                     m_ColumnFormat = m_ColumnFormat.replace("%Y", "%04Y")
@@ -866,6 +974,8 @@ class SQLExecute(object):
                                 elif type(column[m_nPos]) == datetime.datetime:
                                     m_ColumnValue = m_ColumnValue + "TIMESTAMP '" + \
                                                     format_column(column[m_nPos], m_ColumnType) + "'"
+                                elif isinstance(column[m_nPos], type(None)):
+                                    m_ColumnValue = m_ColumnValue + "<null>"
                                 else:
                                     m_ColumnValue = m_ColumnValue + \
                                                     str(format_column(column[m_nPos], m_ColumnType))
@@ -878,6 +988,8 @@ class SQLExecute(object):
                                 elif type(column[m_nPos]) == datetime.datetime:
                                     m_ColumnValue = m_ColumnValue + ",TIMESTAMP '" + \
                                                     format_column(column[m_nPos], m_ColumnType) + "'"
+                                elif isinstance(column[m_nPos], type(None)):
+                                    m_ColumnValue = m_ColumnValue + ",<null>"
                                 else:
                                     m_ColumnValue = m_ColumnValue + "," + \
                                                     str(format_column(column[m_nPos], m_ColumnType))
@@ -885,9 +997,10 @@ class SQLExecute(object):
                         m_row.append(m_ColumnValue)
                     elif columntypes[m_nColumnPos] == "ARRAY":
                         m_ColumnValue = "ARRAY["
-                        if self.SQLOptions.get('OUTPUT_SORT_ARRAY') == "1":
+                        if self.SQLOptions.get('OUTPUT_SORT_ARRAY') == "ON":
                             # 保证Array的输出每次都一样顺序
-                            column.sort()
+                            # 需要注意可能有NULL值导致字符数组无法排序的情况
+                            column.sort(key=lambda x: -1 * float('inf') if x is None else x)
                         for m_nPos in range(0, len(column)):
                             m_ColumnType = str(type(column[m_nPos]))
                             if m_nPos == 0:
@@ -899,6 +1012,8 @@ class SQLExecute(object):
                                 elif type(column[m_nPos]) == datetime.datetime:
                                     m_ColumnValue = m_ColumnValue + "TIMESTAMP '" + \
                                                     format_column(column[m_nPos], m_ColumnType) + "'"
+                                elif isinstance(column[m_nPos], type(None)):
+                                    m_ColumnValue = m_ColumnValue + "<null>"
                                 else:
                                     m_ColumnValue = m_ColumnValue + \
                                                     str(format_column(column[m_nPos], m_ColumnType))
@@ -911,6 +1026,8 @@ class SQLExecute(object):
                                 elif type(column[m_nPos]) == datetime.datetime:
                                     m_ColumnValue = m_ColumnValue + ",TIMESTAMP '" + \
                                                     format_column(column[m_nPos], m_ColumnType) + "'"
+                                elif isinstance(column[m_nPos], type(None)):
+                                    m_ColumnValue = m_ColumnValue + ",<null>"
                                 else:
                                     m_ColumnValue = m_ColumnValue + "," + \
                                                     str(format_column(column[m_nPos], m_ColumnType))
